import typing
import collections.abc

from fundi.types import CallableInfo
from fundi.inject import injection_impl


def tree(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: (
        collections.abc.MutableMapping[
            typing.Callable[..., typing.Any], collections.abc.Mapping[str, typing.Any]
        ]
        | None
    ) = None,
) -> collections.abc.Mapping[str, typing.Any]:
    """
    Get tree of dependencies of callable.

    :param scope: container with contextual values
    :param info: callable information
    :param cache: tree generation cache
    :return: Tree of dependencies
    """
    if cache is None:
        cache = {}

    gen = injection_impl(scope, info, cache, None)

    value = None

    while True:
        inner_scope, inner_info, more = gen.send(value)
        if not more:
            return {"call": inner_info.call, "values": inner_scope}

        value = tree(inner_scope, inner_info, cache)


def order(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: (
        collections.abc.MutableMapping[
            typing.Callable[..., typing.Any], list[typing.Callable[..., typing.Any]]
        ]
        | None
    ) = None,
) -> list[typing.Callable[..., typing.Any]]:
    """
    Get resolving order of callable dependencies.

    :param info: callable information
    :param scope: container with contextual values
    :param cache: solvation cache
    :return: order of dependencies
    """
    if cache is None:
        cache = {}

    gen = injection_impl(scope, info, cache, None)

    order_: list[typing.Callable[..., typing.Any]] = []

    value = None
    while True:
        inner_scope, inner_info, more = gen.send(value)
        if not more:
            return order_

        order_.extend(order(inner_scope, inner_info, cache))
        order_.append(inner_info.call)



import typing
import warnings
import functools

from fundi.scan import scan
from fundi.util import callable_str
from fundi.types import R, DependencyConfiguration

P = typing.ParamSpec("P")


class MutableConfigurationWarning(UserWarning):
    pass


def configurable_dependency(configurator: typing.Callable[P, R]) -> typing.Callable[P, R]:
    """
    Create dependency configurator that caches configured dependencies.
    This helps FunDI cache resolver understand that dependency already executed, if it was.

    Note: Calls with mutable arguments will not be stored in cache and warning would be shown

    :param configurator: Original dependency configurator
    :return: cache aware dependency configurator
    """
    dependencies: dict[frozenset[tuple[str, typing.Any]], R] = {}
    info = scan(configurator)

    if info.async_:
        raise ValueError("Dependency configurator should not be asynchronous")

    @functools.wraps(configurator)
    def cached_dependency_generator(*args: typing.Any, **kwargs: typing.Any) -> R:
        use_cache = True
        values = info.build_values(*args, **kwargs)
        key: frozenset[tuple[str, typing.Any]] | None = None

        try:
            key = frozenset(values.items())

            if key in dependencies:
                return dependencies[key]
        except TypeError:
            warnings.warn(
                f"Can't cache dependency created via {callable_str(configurator)}: configured with unhashable arguments",
                MutableConfigurationWarning,
            )
            use_cache = False

        dependency = configurator(*args, **kwargs)
        setattr(
            dependency,
            "__fundi_configuration__",
            DependencyConfiguration(configurator=info, values=values),
        )

        if use_cache and key is not None:
            dependencies[key] = dependency

        return dependency

    return cached_dependency_generator



import typing
import collections
import collections.abc
from dataclasses import dataclass, field

__all__ = [
    "R",
    "Parameter",
    "TypeResolver",
    "CallableInfo",
    "InjectionTrace",
    "ParameterResult",
    "DependencyConfiguration",
]

R = typing.TypeVar("R")


@dataclass
class TypeResolver:
    """
    Mark that tells ``fundi.scan.scan`` to set ``Parameter.resolve_by_type`` to True.

    This changes logic of ``fundi.resolve.resolve``, so it uses ``Parameter.annotation``
    to find value in scope instead of ``Parameter.name``
    """

    annotation: type


@dataclass
class Parameter:
    name: str
    annotation: typing.Any
    from_: "CallableInfo[typing.Any] | None"
    default: typing.Any = None
    has_default: bool = False
    resolve_by_type: bool = False
    positional_only: bool = False
    keyword_only: bool = False


@dataclass
class CallableInfo(typing.Generic[R]):
    call: typing.Callable[..., R]
    use_cache: bool
    async_: bool
    generator: bool
    parameters: list[Parameter]
    return_annotation: typing.Any
    configuration: "DependencyConfiguration | None"
    named_parameters: dict[str, Parameter] = field(init=False)

    def __post_init__(self):
        self.named_parameters = {p.name: p for p in self.parameters}

    def build_values(self, *args: typing.Any, **kwargs: typing.Any) -> dict[str, typing.Any]:
        arguments: dict[str, typing.Any] = {}

        args_amount = len(args)

        ix = 0
        for parameter in self.parameters:
            name = parameter.name

            if name in kwargs:
                arguments[name] = kwargs[name]
                continue

            if ix < args_amount:
                arguments[name] = args[ix]
                ix += 1
                continue

            if parameter.has_default:
                arguments[name] = parameter.default
                continue

            raise ValueError(f'Argument for parameter "{parameter.name}" not found')

        return arguments

    def build_arguments(
        self, values: collections.abc.Mapping[str, typing.Any]
    ) -> tuple[tuple[typing.Any, ...], dict[str, typing.Any]]:
        positional: tuple[typing.Any, ...] = ()
        keyword: dict[str, typing.Any] = {}

        for name, value in values.items():
            if name not in self.named_parameters:
                raise ValueError(f'Parameter named "{name}" not found')

            parameter = self.named_parameters[name]

            if parameter.positional_only:
                positional += (value,)
            elif parameter.keyword_only:
                keyword[name] = value
            else:
                positional += (value,)

        return positional, keyword


@dataclass
class ParameterResult:
    parameter: Parameter
    value: typing.Any | None
    dependency: CallableInfo[typing.Any] | None
    resolved: bool


@dataclass
class InjectionTrace:
    info: CallableInfo[typing.Any]
    values: collections.abc.Mapping[str, typing.Any]
    origin: "InjectionTrace | None" = None


@dataclass
class DependencyConfiguration:
    configurator: CallableInfo[typing.Any]
    values: collections.abc.Mapping[str, typing.Any]



import typing
from typing import overload
from collections.abc import Awaitable, Iterable, AsyncIterable

T = typing.TypeVar("T", bound=type)
R = typing.TypeVar("R")

@overload
def from_(dependency: T, caching: bool = True) -> T: ...
@overload
def from_(dependency: typing.Callable[..., Iterable[R]], caching: bool = True) -> R: ...
@overload
def from_(dependency: typing.Callable[..., AsyncIterable[R]], caching: bool = True) -> R: ...
@overload
def from_(dependency: typing.Callable[..., Awaitable[R]], caching: bool = True) -> R: ...
@overload
def from_(dependency: typing.Callable[..., R], caching: bool = True) -> R: ...



import typing

from fundi.util import callable_str
from fundi.types import CallableInfo


class ScopeValueNotFoundError(ValueError):
    def __init__(self, parameter: str, info: CallableInfo[typing.Any]):
        super().__init__(
            f'Cannot resolve "{parameter}" for {callable_str(info.call)} - Scope does not contain required value'
        )
        self.parameter: str = parameter
        self.info: CallableInfo[typing.Any] = info



import typing
import collections.abc
from contextlib import ExitStack, AsyncExitStack

from fundi.resolve import resolve
from fundi.types import CallableInfo
from fundi.util import call_sync, call_async, add_injection_trace


def injection_impl(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: collections.abc.MutableMapping[typing.Callable[..., typing.Any], typing.Any],
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None,
) -> collections.abc.Generator[
    tuple[collections.abc.Mapping[str, typing.Any], CallableInfo[typing.Any], bool],
    typing.Any,
    None,
]:
    """
    Injection brain.

    Coordinates dependency resolution for a given `CallableInfo`. For each parameter:

    - If the parameter has a pre-resolved value (from scope, override, or cache) — uses it.
    - If the parameter requires another dependency to be resolved:
      - Yields `(scope_with_context, dependency_info, True)` to request the caller to inject it.
      - Once the value is received — caches it if allowed.

    After all parameters are resolved, yields:
      `(resolved_values_dict, top_level_callable_info, False)`

    If any error occurs during resolution, attaches injection trace and re-raises the exception.
    """

    values: dict[str, typing.Any] = {}
    try:
        for result in resolve(scope, info, cache, override):
            name = result.parameter.name
            value = result.value

            if not result.resolved:
                dependency = result.dependency
                assert dependency is not None

                value = yield {**scope, "__fundi_parameter__": result.parameter}, dependency, True

                if dependency.use_cache:
                    cache[dependency.call] = value

            values[name] = value

        yield values, info, False

    except Exception as exc:
        add_injection_trace(exc, info, values)
        raise exc


def inject(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: ExitStack,
    cache: (
        collections.abc.MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None
    ) = None,
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> typing.Any:
    """
    Synchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :param override: override dependencies
    :return: result of callable
    """
    if info.async_:
        raise RuntimeError("Cannot process async functions in synchronous injection")

    if cache is None:
        cache = {}

    gen = injection_impl(scope, info, cache, override)

    value: typing.Any | None = None

    try:
        while True:
            inner_scope, inner_info, more = gen.send(value)

            if more:
                value = inject(inner_scope, inner_info, stack, cache, override)
                continue

            return call_sync(stack, inner_info, inner_scope)
    except Exception as exc:
        add_injection_trace(exc, info, scope)
        raise exc


async def ainject(
    scope: collections.abc.Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    stack: AsyncExitStack,
    cache: (
        collections.abc.MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None
    ) = None,
    override: collections.abc.Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> typing.Any:
    """
    Asynchronously inject dependencies into callable.

    :param scope: container with contextual values
    :param info: callable information
    :param stack: exit stack to properly handle generator dependencies
    :param cache: dependency cache
    :param override: override dependencies
    :return: result of callable
    """
    if cache is None:
        cache = {}

    gen = injection_impl(scope, info, cache, override)

    value: typing.Any | None = None

    try:
        while True:
            inner_scope, inner_info, more = gen.send(value)

            if more:
                value = await ainject(inner_scope, inner_info, stack, cache, override)
                continue

            if info.async_:
                return await call_async(stack, inner_info, inner_scope)

            return call_sync(stack, inner_info, inner_scope)
    except Exception as exc:
        add_injection_trace(exc, info, scope)
        raise exc



import typing
from typing import overload
from contextlib import ExitStack as SyncExitStack, AsyncExitStack
from collections.abc import Generator, AsyncGenerator, Mapping, MutableMapping, Awaitable

from fundi.types import CallableInfo

R = typing.TypeVar("R")

ExitStack = AsyncExitStack | SyncExitStack

def injection_impl(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[typing.Any],
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any],
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None,
) -> Generator[
    tuple[Mapping[str, typing.Any], CallableInfo[typing.Any], bool],
    typing.Any,
    None,
]: ...
@overload
def inject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[Generator[R, None, None]],
    stack: ExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...
@overload
def inject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[R],
    stack: ExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...
@overload
async def ainject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[Generator[R, None, None]],
    stack: AsyncExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...
@overload
async def ainject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[AsyncGenerator[R, None]],
    stack: AsyncExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...
@overload
async def ainject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[Awaitable[R]],
    stack: AsyncExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...
@overload
async def ainject(
    scope: Mapping[str, typing.Any],
    info: CallableInfo[R],
    stack: AsyncExitStack,
    cache: MutableMapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
    override: Mapping[typing.Callable[..., typing.Any], typing.Any] | None = None,
) -> R: ...



import typing
import inspect
import warnings
import collections.abc
from types import TracebackType
from contextlib import AsyncExitStack, ExitStack

from fundi.types import CallableInfo, InjectionTrace, DependencyConfiguration


__all__ = [
    "call_sync",
    "call_async",
    "callable_str",
    "is_configured",
    "injection_trace",
    "get_configuration",
    "add_injection_trace",
]


def callable_str(call: typing.Callable[..., typing.Any]) -> str:
    if hasattr(call, "__qualname__"):
        name = call.__qualname__
    elif hasattr(call, "__name__"):
        name = call.__name__
    else:
        name = str(call)

    module = inspect.getmodule(call)

    module_name = "<unknown>" if module is None else module.__name__

    return f"<{name} from {module_name}>"


def add_injection_trace(
    exception: Exception,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> None:
    setattr(
        exception,
        "__fundi_injection_trace__",
        InjectionTrace(info, values, getattr(exception, "__fundi_injection_trace__", None)),
    )


def call_sync(
    stack: ExitStack | AsyncExitStack,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> typing.Any:
    """
    Synchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    args, kwargs = info.build_arguments(values)
    value = info.call(*args, **kwargs)

    if info.generator:
        generator: collections.abc.Generator[typing.Any, None, None] = value
        value = next(generator)

        def close_generator(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                if exc_type is not None:
                    generator.throw(exc_type, exc_value, tb)
                else:
                    next(generator)
            except StopIteration:
                # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
                return exc_type is None
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            warnings.warn("Generator not exited", UserWarning)

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push(close_generator)

    return value


async def call_async(
    stack: AsyncExitStack,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> typing.Any:
    """
    Asynchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    args, kwargs = info.build_arguments(values)

    value = info.call(*args, **kwargs)

    if info.generator:
        generator: collections.abc.AsyncGenerator[typing.Any] = value
        value = await anext(generator)

        async def close_generator(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                if exc_type is not None:
                    await generator.athrow(exc_type, exc_value, tb)
                else:
                    await anext(generator)
            except StopAsyncIteration:
                # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
                return exc_type is None
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            warnings.warn("Generator not exited", UserWarning)

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push_async_exit(close_generator)

    else:
        value = await value

    return value


def injection_trace(exception: Exception) -> InjectionTrace:
    """
    Get injection trace from exception

    :param exception: exception to get injection trace from
    :return: injection trace
    """
    if not hasattr(exception, "__fundi_injection_trace__"):
        raise ValueError(f"Exception {exception} does not contain injection trace")

    return typing.cast(InjectionTrace, getattr(exception, "__fundi_injection_trace__"))


def is_configured(call: typing.Callable[..., typing.Any]) -> bool:
    """
    Get whether callable is configured via @configurable_dependency

    :param call: callable to check
    :return: Is this callable configured
    """
    return hasattr(call, "__fundi_configuration__")


def get_configuration(call: typing.Callable[..., typing.Any]) -> DependencyConfiguration:
    """
    Get dependency configuration. Can be useful in third-party tools that needs to know configuration

    :param call: callable to get configuration from
    :return: dependency configuration
    """
    if not is_configured(call):
        raise ValueError(f"Callable {call} is not configured via @configurable_dependency")

    configuration: DependencyConfiguration = getattr(call, "__fundi_configuration__")
    return configuration



import typing
import inspect

from fundi.util import is_configured, get_configuration
from fundi.types import R, CallableInfo, Parameter, TypeResolver


def scan(call: typing.Callable[..., R], caching: bool = True) -> CallableInfo[R]:
    """
    Get callable information

    :param call: callable to get information from
    :param caching:  whether to use cached result of this callable or not

    :return: callable information
    """
    params: list[Parameter] = []
    signature = inspect.signature(call)

    for param in signature.parameters.values():
        positional_only = param.kind == inspect.Parameter.POSITIONAL_ONLY
        keyword_only = param.kind == inspect.Parameter.KEYWORD_ONLY
        if isinstance(param.default, CallableInfo):
            params.append(
                Parameter(
                    param.name,
                    param.annotation,
                    from_=typing.cast(CallableInfo[typing.Any], param.default),
                    positional_only=positional_only,
                    keyword_only=keyword_only,
                )
            )
            continue

        has_default = param.default is not inspect.Parameter.empty
        resolve_by_type = False

        annotation = param.annotation
        if isinstance(annotation, TypeResolver):
            annotation = annotation.annotation
            resolve_by_type = True

        elif typing.get_origin(annotation) is typing.Annotated:
            args = typing.get_args(annotation)

            if args[1] is TypeResolver:
                resolve_by_type = True

        params.append(
            Parameter(
                param.name,
                annotation,
                from_=None,
                default=param.default if has_default else None,
                has_default=has_default,
                resolve_by_type=resolve_by_type,
                positional_only=positional_only,
                keyword_only=keyword_only,
            )
        )

    async_: bool = inspect.iscoroutinefunction(call) or inspect.isasyncgenfunction(call)
    generator: bool = inspect.isgeneratorfunction(call) or inspect.isasyncgenfunction(call)

    info = typing.cast(
        CallableInfo[R],
        CallableInfo(
            call=call,
            use_cache=caching,
            async_=async_,
            generator=generator,
            parameters=params,
            return_annotation=signature.return_annotation,
            configuration=get_configuration(call) if is_configured(call) else None,
        ),
    )

    return info
